/**
Experimental code for the ncurses bindings.
**/

if( ! is_defined( 'ncurses' ) )
{
	var dll = '../ncurses.so';
	if( ! open_dll(dll) ) {
		throw new Error("JSWindow could not load plugin '"+dll+"'.");
	}
}

if( ! is_defined('JCWindow') ) {
	assert(include('inheritance.inc.js'));

	/**
	   JCWindow is an OO ncurses Window class based on top of the
	   nc_XXX() windowing API. It is intended to replace
	   the NCWindow family of classes.

	   Creating a JCWindow() before nc_initscr() has been called
	   will force nc_initscr() to be called.

	   Prerequisites:

	   - The ncurses plugin must be loaded.


	   JCWindow ctors:

	   JCWindow(): a wrapper object for ncurses' built-in stdscr.
	   All JCWindows created this way point to the same stdscr
	   window.

	   JCWindow(lines,cols,y,x): creates a panelized top-level window.
	   
	   JCWindow(JCWindow parent): makes a new subwindow of parent
	   which is two rows/cols smaller than parent.
	   
	   JCWindow(JCWindow parent,lines,cols,y,x): makes a new
	   subwindow with the given geometry.

	   JCWindow(WINDOW|PANEL): makes the JCWindow a wrapper around the
	   given WINDOW.

	   Public properties:

	   .window = a WINDOW handle. Don't reassign it.

	   .panel = a PANEL handle ONLY if this object is created with
	   JCWindow(PANEL) or JCWindow(l,c,y,x), else it is null.

	   .pad = Normally null, but sublcasses might set it
	   to a non-null value (an ncurses PAD handle). If
	   this object has a .pad, then .destroy() will also
	   destroy that pad.


	   Signals:

	   You can connect Slot objects to the Signals contained
	   by this class in order to be made aware of certain
	   actions. To connect to a Signal:

	   var myslot = myJSWindow.signals.SignalName.connect( new Slot( my_function ) )

	   If you do not "root" the new slot by assigning it to
	   something then it may disappear before you are done with it
	   (it may get prematurely garbage collected).

	   See the Signal/Slot API docs for more details.

	   This class fires the following Signals:

	   .signals.[nout]refresh.fire(JCWindow) is called just AFTER
	   a [nout]refresh() is executed. Slots can, e.g., refresh
	   related windows when this is called.

	   .signals.getch.fire(JCWindow,keycode) is called after a getch()
	   is executed.

	   .signals.[show|hide].fire(JCWindow,bool) is called for
	   PANELs when show() or hide() is called. The bool value
	   specifies whether the panel is currently shown or hidden.

	*/
	function JCWindow() {
		
		if( ! JCWindow.inited )
		{
			JCWindow.inited = true;
			if( ! is_defined( 'stdscr' ) )
			{
				if( !nc_initscr() )
				{
					throw new Error("JCWindow(): nc_initscr() failed!");
				}
			}
			nc_meta(stdscr,true);
			nc_cbreak();
			nc_noecho();
		}
		var self = this;
		this.is_a_JCWindow = true;
		this.signals = new Object();
		// TODO: move these member funcs to the class prototype!
		////////////////////////////////////////////////////////////////////////
		// Define public member functions...
		// More are defined outside of this function via the Function.method(...)
		// technique
		this.touch = function() { return nc_touchwin(this.window); };
		this.touchline = function(line,count,changed) { return nc_wtouchln(this.window,line,count,changed); };
		this.keypad = function(b) { return nc_keypad(this.window,b); };
		this.getch = function() {
			var c = nc_wgetch(self.window);
			self.signals.getch.fire(self,c);
			return c;
		};
		this.signals.getch = new Signal();
		this.refresh = function() {
			var rc = 0;
			if( this.pad ) return -1;
// 			if( this.panel )
// 			{
// 				rc = nc_update_panels();
// 			}
// 			else
// 			{
 				rc = nc_wrefresh(self.window);
// 			}
			self.signals.refresh.fire(self);
			return rc;
		};
		this.signals.refresh = new Signal();
		//this.signals.refresh.debug(true);

		this.noutrefresh = function() {
			var rc = nc_wnoutrefresh(self.window);
			self.signals.noutrefresh.fire(self);
			return rc;
		};
		this.signals.noutrefresh = new Signal();

		this.update = function() {
			this.refresh();
			nc_doupdate();
		};

		////////////////////////////////////////////////////////////////////////
		// End functions and start with ctor logic...
		self.window = null;
		self.panel = null;
		self.pad = null;
		self.parent = null;
		var arg0 = arguments[0];
		if( 0 == arguments.length )
		{ // JCWindow == stdscr
			self.window = stdscr;
		}
		else if( 1 == arguments.length )
		{
			var p = arguments[0];
			if( nc_is_pad(p) )
			{ // JCWindow(PAD)
				//self.window = self.pad = p;
				// Not handled here
			}
			else if( nc_is_panel(p) )
			{ // JCWindow(PANEL)
				self.panel = p;
				self.window = nc_panel_window(p);
				//self.window = p;
			}
			else if( nc_is_window(p) )
			{ // JCWindow(WINDOW)
				self.window = p;
			}
			else if( is_object(p) && is_defined( p, 'is_a_JCWindow' ) )
			{ // assume JCWindow(JCWindow parent)
				var args = new Array();
				args[0] = p.window;
				args[1] = p.height()-2;
				args[2] = p.width()-2;
				args[3] = 1;
				args[4] = 1;
				self.window = nc_derwin.apply(self,args);
				if( ! self.window )
				{
					throw new Error("JCWindow ctor: nc_derwin() failed!");
				}
				self.parent = p;
			}
		}
		else if( 4 == arguments.length )
		{ // JCWindow(l,c,y,x)
			self.window = nc_newwin.apply(self,JCWindow.copyargs(arguments));
			if( self.window )
			{
				self.panel = //nc_new_panel.apply(self.window);
					nc_new_panel(self.window);
			}

		}
		else if( 5 == arguments.length )
		{//JCWindow(JCWindow parent,l,c,y,x)
			//cerr( 'JCWindow(argc==',arguments.length,', args==[',JCWindow.copyargs(arguments).toString(),'])');
			var par = arguments[0];
			if( (!is_object(par))
			    &&
			    (!is_defined( par, 'is_a_JCWindow' ))
			    )
			{
				throw new Error("JCWindow() ctor: 5-arg ctor expects a (JCWindow) as the first argument.");
			}
			var args = new Array();
			args[0] = par.window;
			//print("p.window ==",p.window);
			for( var i = 1; i < 5; ++i ) args[i] = arguments[i];
			self.window = nc_derwin.apply(self,args);
			if( ! self.window )
			{
				throw new Error('JCWindow(): nc_derwin(',args.toSource(),') failed.');
			}
			self.parent = par;
		}
		
		if( ! self.window )
		{
			throw new Error( "JCWindow ctor: unhandled arguments or window init failed. Read the API docs." );
		}

		if( self.panel )
		{
			self.show = function() {
			//nc_show_panel.apply(arguments,[0]);
 				var s = (arguments.length ? arguments[0] : true);
 				if( s ) nc_top_panel(self.panel); // kludge!
 				var rc = nc_show_panel(self.panel, s);
				self.signals.show.fire(self,s);
				return rc;
			};
			self.signals.show = new Signal();

			self.hide = function() {
				var s = (arguments.length ? arguments[0] : true);
				var rc = nc_hide_panel(self.panel,s);
				self.signals.show.fire(self,!s);
				return rc;
			};
			self.hidden = function() {
				return nc_panel_hidden(self.panel) ? true : false;
			};
			self.top = function() { nc_top_panel(self.panel); };
			self.bottom = function() { nc_bottom_panel(self.panel); };
			self.refresh = function() {
				//var rc = nc_wrefresh(self.window);
				var rc = nc_update_panels();
				self.signals.refresh.fire(self);
				return rc;
			};
		}
		self.meta(true);
		self.keypad(true);
		return self;

	} // JCWindow
	JCWindow.inited = false; // internal static init var
	/**
	   JCWindow.copyargs is an internal convenience function.
	   Pass it your 'arguments' semi-array and it will create
	   a real array out of it.
	*/
	JCWindow.copyargs = function(arg) {
		var a = new Array();
		for( var i = 0; i < arg.length; ++i ) { a[i] = arg[i]; }
		return a;
	};

 	JCWindow.method('background',
			function(b) {
				if( is_void(b) )
				{
					return nc_getbkgd(this.window);
				}
				return nc_wbkgd(this.window,b);
			} );
 	JCWindow.method('height', function() { return nc_wheight(this.window); } );
 	JCWindow.method('meta', function(b) { return nc_meta(this.window,b); } );
	/**
	   movecursor(y,x[,relative=false]). Tries to move the cursor
	   to the requested y/x coordinate. If relative is true then
	   the coordinates are considered to be relative offsets from
	   the current coordinates.
	*/
 	JCWindow.method('movecursor', function(y,x,relative) {
				if( is_void(relative) ) relative = false;
				if( relative ) {
					y += this.cursory();
					x += this.cursorx();
				}
				return nc_wmove(this.window,y,x);
			} );
	/**
	   movewin(y,x[,relative=false]). Tries to move the window
	   to the requested y/x coordinate. If relative is true then
	   the coordinates are considered to be relative offsets from
	   the current position.

	   Note that moving parented windows might have undesirable
	   results.
	*/
 	JCWindow.method('movewin',
			function(y,x,relative)
			{
				if( is_void(relative) ) relative = false;
				if( relative ) {
					y += this.winy();
					x += this.winx();
				}
				if( this.panel )
				{
					return nc_move_panel(this.panel,y,x);
				}
				return nc_movewin(this.window,y,x);
			} );

 	JCWindow.method('scrollok', function(b) { nc_scrollok( this.window, b ); } );
 	JCWindow.method('width', function() { return nc_wwidth(this.window); } );
 	JCWindow.method('winx', function() { return nc_wbegx(this.window); } );
 	JCWindow.method('winy', function() { return nc_wbegy(this.window); } );
 	JCWindow.method('cursorx', function() { return nc_getcurx(this.window); } );
 	JCWindow.method('cursory', function() { return nc_getcury(this.window); } );
	JCWindow.method('toString',
			function() {
				if( is_object( this.window ) ) return this.window.toString();
				return ""+this.window;
			} );
	JCWindow.method('box', function() {
				return nc_box(this.window);
			} );

	JCWindow.method('border', function() {
			var a = new Array();
			a[0] = this.window;
			for( var i = 0; i<arguments.length; ++i )
			{
				a[1+i] = arguments[i];
			}
			return nc_wborder.apply( this.window, Array.prototype.slice.apply( a, [0] ) );
			} );

	JCWindow.method('attron',
			function() {
				return nc_wattron(this.window,arguments[0]);
			} );
	JCWindow.method('attroff',
			function() {
				return nc_wattroff(this.window,arguments[0]);
			} );
	JCWindow.method('attrset',
			function() {
				return nc_wattrset(this.window,arguments[0]);
			} );
	JCWindow.method('clear',
			function() {
				return nc_wclear(this.window);
			} );
	JCWindow.method('erase',
			function() {
				return nc_werase(this.window);
			} );

	JCWindow.method('addch',
			function() {
				var arg = JCWindow.copyargs(arguments);
				if( 1 == arg.length ) {
					// (char)
					return nc_waddch( this.window, arg[0] );
				}
				else if( 3 == arg.length ) {
					// (y,x,char)
					return nc_mvwaddch(this.window,arg[0], arg[1], arg[2]);
				}
				else {
					throw new Error("JCWindow::addch() requires either (char) or (y,x,char) arguments.");
				}
				return -1; // can't happen.
			} );
	JCWindow.method('addstr',
			function() {
				var arg = JCWindow.copyargs(arguments);
				if( 1 == arg.length ) {
					// (string)
					return nc_waddstr( this.window, arg[0] );
				}
				else if( 2 == arg.length ) {
					// (string,length)
					return nc_waddnstr(this.window,arg[0],arg[1]);
				}
				else if( 3 == arg.length ) {
					// (y,x,string)
					return nc_mvwaddstr(this.window,arg[0],arg[1],arg[2]);
				}
				else if( 4 == arg.length ) {
					// (y,x,string,length)
					return nc_mvwaddnstr(this.window,arg[0],arg[1],arg[2],arg[3]);
				}
				throw new Error("addstr(): unhandled arguments: "+arguments.toSource());
			} );
	JCWindow.method('addvstr',
			function() {
				var arg = JCWindow.copyargs(arguments);
				if( 1 == arg.length ) {
					// (string)
					return nc_waddvstr( this.window, arg[0] );
				}
				else if( 2 == arg.length ) {
					// (string,length)
					return nc_waddvnstr(this.window,arg[0],arg[1]);
				}
				else if( 3 == arg.length ) {
					// (y,x,string)
					return nc_mvwaddvstr(this.window,arg[0],arg[1],arg[2]);
				}
				else if( 4 == arg.length ) {
					// (y,x,string,length)
					return nc_mvwaddvnstr(this.window,arg[0],arg[1],arg[2],arg[3]);
				}
				throw new Error("addvstr(): unhandled arguments: "+arguments.toSource());
			} );

	JCWindow.method('destroy',
			function() {
				if( stdscr == this.window ) {
					nc_endwin();
					this.window = this.panel = this.pad = null;
					return;
				}
				if( this.panel ) {
					this.hide();
					nc_del_panel( this.panel );
				}
				if( this.pad != this.window )
				{
					nc_delwin(this.pad);
				}
				nc_delwin(this.window);
				this.window = this.panel = null;
			} );

 	JCWindow.method('hline',
			/**
			   args must be one of:

			   (char,count)

			   (y,x,char,count)
			*/
			function()
			{
				var args = JCWindow.copyargs(arguments);
				if( 2 == args.length )
				{ // (chtype,count)
					return nc_whline( this.window, args[0], args[1] );
				}
				else if( 4 == args.length )
				{ // (y, x, chtype, count)
					return nc_mvwhline( this.window, args[0], args[1], args[2], args[3] );
				}
				return -1;
			} );

 	JCWindow.method('vline',
			/**
			   args must be one of:

			   (char,count)

			   (y,x,char,count)
			*/
			function()
			{
				var args = JCWindow.copyargs(arguments);
				if( 2 == args.length )
				{ // (chtype,count)
					return nc_wvline( this.window, args[0], args[1] );
				}
				else if( 4 == args.length )
				{ // (y,x,chtype,count)
					return nc_mvwvline( this.window, args[0], args[1], args[2], args[3] );
				}
				return -1;
			} );




	/**

	Constructor for a JCPad, which acts as a wrapper for an
	ncurses PAD object.

	Constructors:

	- JCPad(width,height): creates a new ncurses PAD of the given
	size.

	- JCPad(JCPad | PAD): this object becomes a wrapper for the
	given pad.

	Postconditions:

	this.window == this.pad

	This type subclasses JCWindow, and inherits its functions, but
	note that not all ncurses WINDOW operations are legal on PADs,
	e.g. nc_wrefresh().

	*/
	function JCPad()
	{
		var argc = arguments.length;
		var self = this;
		this.window = this.panel = this.pad = null;
		this.is_a_JCPad = true;
		if( 2 == argc )
		{ // JCPad(height,width)
			self.pad = nc_newpad( arguments[0],
					      arguments[1] );

			if( ! self.pad )
			{
				throw new Error("JCPad("+arguments[0]+","+arguments[1]+") ctor couldn't create pad." );
			}
		}
		else if( 1 == argc )
		{ // JCPad(JCPad) or JCPad(PAD)
			var a = arguments[0];
			var p = a;
			if( is_object( a ) && is_defined( a, 'is_a_JCPad' ) )
			{
				p = p.pad;
			}
			if( nc_is_pad( p ) )
			{
				this.pad = p;
			}
		}

		if( ! self.pad )
		{
			throw new Error("JCPad() unhandled ctor arguments. Read the docs." );
		}

		self.window = self.pad; // is this sane?

		return self;
	}

	JCPad.inherits(JCWindow);

	/** Reimplemented to enforce pad refresh requirements. */
	JCPad.method('refresh', function(y1,x1,y2,x2,y3,x3) {
			     //cerr("PAD REFRESH!",(new Array(y1,x1,y2,x2,y3,x3)).toString());
			     if( 6 != arguments.length )
			     {
				     return -1;
				     //throw new Error("JCPad.refresh(): this object is a PAD, so you must pass 6 coordinate arguments. See 'man prefresh' for details.");
			     }
			     var rc = nc_prefresh(this.pad,y1,x1,y2,x2,y3,x3);
			     this.signals.refresh.fire(this,y1,x1,y2,x2,y3,x3);
			     return rc;
		     } );

	/** Reimplemented to enforce pad refresh requirements. */
	JCPad.method('noutrefresh', function(y1,x1,y2,x2,y3,x3) {
			     //cerr("PAD NOUTREFRESH!",(new Array(y1,x1,y2,x2,y3,x3)).toString());
			     if( 6 != arguments.length )
			     {
				     return -1;
				     //throw new Error("JCPad.noutrefresh(): this object is a PAD, so you must pass 6 coordinate arguments. See 'man pnoutrefresh' for details.");
			     }
			     var rc = nc_pnoutrefresh(this.pad,y1,x1,y2,x2,y3,x3);
			     this.signals.noutrefresh.fire(this,y1,x1,y2,x2,y3,x3);
			     return rc;
		     } );


	/**
	   A JCScroller is a wrapper around an ncurses PAD and
	   a WINDOW. It subclasses JCWindow, so see that class
	   for more details, like additional properties.

	   This type isn't yet suitable for use with overlapping
	   panels - it has some refresh problems with those.

	   Constructor arguments:

	   - JCScroller( PAD | JCPad, WINDOW | JCWindow ): Creates
	   a wrapper around the given pad and window. This object
	   simplifies the process of updating the window ("viewport")
	   with the contents of the pad.

	   - JCScroller( lines, cols, y, x, scrollines, scrollcols ):
	   creates a new top-level paneled window. The
	   scrolllines/cols must be bigger than the window lines/cols
	   or display problems might arise.

	   Properties:

	   .pad and .window are the same object: the PAD ctor argument
	   or the PAD created for the 6-arg ctor.

	   .viewport is the second ctor argument. This is where
	   refreshing is done.

	   Signals:

	   this.signals.scrollto(this,result_code) is fired after
	   this.scrollto() is called. See that function for details.

	   Notes:

	   Not all window operations are legal or sensible (or
	   predictable) on pads. You may have to experiment.

	   This class reimplements the inherited refresh() method
	   to update this.viewport on a refresh. This removes
	   the client's need to know whether he's refreshing a
	   pad or a window.

	*/
	function JCScroller()
	{
		var argc = arguments.length;
		var self = this;
		this.window = null;
		this.pad = null;
		this.is_a_JCScroller = true;
		this.viewport = null;
		this.spos = {y:-1,x:-1};
		if( 2 == argc )
		{ // JCScroller(PAD,WINDOW)
			var p = arguments[0];
			var w = arguments[1];
			if( is_defined( p, 'is_a_JCPad' ) ) {
				p = p.pad;
			}
			if( is_defined( w, 'is_a_JCWindow' ) ) {
				w = w.window;
			}
			if( nc_is_pad( p )
			    &&
			    nc_is_window( w )
			    ) {
				self.window = self.pad = p;
				self.viewport = w;
			}
			if( ! self.window )
			{
				throw new Error("JCScroller(PAD,WINDOW) ctor failed: argument(s) not of correct type(s)." );
			}
		} else if( 6 == argc )
		{
			var pos = 0;
			var l = arguments[pos++];
			var c = arguments[pos++];
			var y = arguments[pos++];
			var x = arguments[pos++];
			var sl = arguments[pos++];
			var sc = arguments[pos++];
			self.Window = new JCPad(sl,sc);
			//self.Vindow = new JCWindow(l,c,y,x);
			self.window = self.pad = Pad.pad;
			self.viewport = Window.window
		}

		if( ! self.viewport )
		{
			throw new Error("JCPad() unhandled ctor arguments. Read the docs." );
		}

// 		this.vheight = function() { return nc_wheight( self.viewport ); };
// 		this.vwidth = function() { return nc_wwidth( self.viewport );};
// 		this.pheight = function() { return nc_wheight( self.pad ); };
// 		this.pwidth = function() { return nc_wwidth( self.pad ); };

		this.signals.scrollto = new Signal();
		return self;

	} // JCScroller

	JCScroller.inherits( JCWindow );

	JCScroller.prototype.vheight = function() { return nc_wheight( this.viewport ); };
	JCScroller.prototype.vwidth = function() { return nc_wwidth( this.viewport );};
	JCScroller.prototype.pheight = function() { return nc_wheight( this.pad ); };
	JCScroller.prototype.pwidth = function() { return nc_wwidth( this.pad ); };

	/**
	   Draws this.pad to the viewport window. If y and x are
	   passed, the top/left position of the viewport will be that
	   y/x position in this.pad. If y and x are not possed, the
	   last scrollto()d position is used.

	   This method updates the viewport window as well. It returns
	   true if it thinks it actually scrolled, else false.

	   If y or x are out of bounds then they are silently moved
	   in bounds. To avoid ugly refresh problems, this algorithm
	   only scrolls to areas where the whole viewport can be filled.
	   e.g., if you tell it to scroll to the last row, it will actually
	   scroll a few rows up (adjusted by the hight of the viewport).

	   After performing the scroll, The Signal
	   this.signals.scrollto(this,rc) is fired ONLY if scrolling
	   actually caused the viewport to move, where rc is the
	   result of the scrollto call.
	*/
	JCScroller.method('scrollto', function(y,x) {
				  if( is_void(y) ) y = this.spos.y;
				  if( is_void(x) ) x = this.spos.x;

// 				  if( -1 == y ) y = this.vheight();
// 				  if( -1 == x ) x = this.vwidth();

				  var oldpos = {y:this.spos.y,
						x:this.spos.x};
				  var pht = nc_wheight(this.pad);
				  var pwt = nc_wwidth(this.pad);

				  var vp = this.viewport;
				  var vht = nc_wheight(vp);
				  var vwt = nc_wwidth(vp);
				  var vy = nc_wbegy(vp);
				  var vx = nc_wbegx(vp);

				  if( y >= (pht-vht) )
				  {
					  y = pht-vht;
				  }
				  if( x >= (pwt-vwt) )
				  {
					  x = pwt-vwt;
				  }

				  //print("pht=",pht,"pwt=",pwt);
				  if( y > pht-2 ) y = pht-2;
				  else if( y <0  ) y = 0;
				  if( x > pwt-2 ) x = pwt-2;
				  else if( x <0  ) x = 0;

				  this.spos.y = y;
				  this.spos.x = x;

				  var rc = false;

				  if( 0 ) {
					  var y2 = vy+vht-1;
					  var x2 = vx+vwt-1;
					  if( 0 ) {
						  // ugly kludge to properly unpaint trailing lines
						  // when the end of the list is in view.
						  nc_wbkgd( vp, nc_getbkgd( this.pad ) );
						  nc_wclear( vp );
						  nc_wnoutrefresh(vp);
					  }
					  rc = (0 == nc_pnoutrefresh( this.pad, y, x, vy, vx, y2, x2 ));
					  nc_update_panels();
					  nc_doupdate();
				  } else {
					  var y2 = vht-1;
					  var x2 = vwt-1;
					  rc = (0 == nc_copywin( this.pad, vp, y, x, 0, 0, y2, x2, 0 ));
					  if( rc )
					  {
						  //rc = nc_wnoutrefresh(this.pad);
						  nc_wnoutrefresh(vp);
						  //nc_wrefresh(vp);
						  // ^^^ doesn't interact well with panels
						  //rc = nc_redrawwin(vp);
						  //nc_update_panels();
						  //nc_doupdate(); // i hate this!
					  }
				  }
				  if( (oldpos.x != this.spos.x)
				      || (oldpos.y != this.spos.y)
				      )
				  {
					  this.signals.scrollto.fire(this,rc);
				  }
				  return rc;
			  } );

	/**
	   Like scrollto(y,x), but scrolls by a relative amount.
	   If either arg is void then 0 is assumed, which means
	   no scrolling will happen, at least not visually.
	*/
	JCScroller.method('scrollby', function(y,x) {
				  if( is_void(y) ) y = 0;
				  if( is_void(x) ) x = 0;
				  return this.scrollto( this.spos.y + y, this.spos.x + x );
			  } );

	/**
	   If key is one of the following ncurses members:

	   PGUP, PGDN, LEFT, RIGHT, DOWN, UP, HOME, END

	   then this function scrolls the viewport by that logical
	   amount and returns the result of this.scrollto(), otherwise
	   it has no side effects and returns -1.

	   If mod is set then it is used as a multiplier, such that
	   LEFT with (mod==3) will scroll 3 to the left.
	*/
	JCScroller.method('apply_key',
			  function(key,mod) {
				  var y = 0;
				  var x = 0;
				  if( is_void(mod) || !mod ) mod = 1;
				  var a = new Array();
				  a[ncurses.PGUP] = {y:-this.vheight(),x:0};
				  a[ncurses.PGDN] = {y:this.vheight(),x:0};
				  a[ncurses.LEFT] = {y:0,x:-1};
				  a[ncurses.RIGHT] = {y:0,x:+1};
				  a[ncurses.DOWN] = {y:+1,x:0};
				  a[ncurses.UP] = {y:-1,x:0};
				  a[ncurses.HOME] = {y:-this.spos.y,x:-this.spos.x};
				  a[ncurses.END] = {y:this.pheight(),x:0};
				  var k = a[key];
				  if( k ) {
					  x = k.x;
					  y = k.y;
				  }
				  delete a;
				  if( x || y )
				  {
					  return this.scrollby( y*mod, x*mod );
				  }
				  return -1;
			  } );
	/**
	   Reimplemented to update the viewport. It bases the view
	   upon the last scrollto() position. Note that this might
	   cause this.signals.scrollto to be fired as well as a
	   this.signals.refresh.
	*/
	JCScroller.method( 'refresh', function() {
				   var rc = this.scrollby( 0, 0 );
				   //var rc = this.scrollto( nc_getcury(this.pad), nc_getcurx(this.pad) );
				   this.signals.refresh.fire( this );
				   return rc;
			  } );
	/**
	   Identical to this.refresh() except that
	   this.signals.noutrefresh is fired instead of
	   this.signals.refresh.
	*/
	JCScroller.method( 'noutrefresh', function() {
				   var rc = this.scrollby( 0, 0 );
				   this.signals.noutrefresh.fire( this );
				   return rc;
			  } );

	/**
	   Tries to render obj to win. The separator parameter is
	   appended to win via nc_addstr() if it is set.

	   This function provides a uniform interface for rendering
	   a variety of types to a WINDOW.

	   Arguments requirements:

	   win must be an ncurses WINDOW handle or a JCWindow
	   instance.

	   obj may be:

	   - non-object: nc_waddstr(win,obj+separator).

	   - a Function object: this function calls obj(win), adds
	   the separator (if any) and returns the value from obj(win).

	   - array: recursively calls nc_wrender() on each array value.
	   If any call fails, the loop stops and the error code is
	   returned. If separator is a string then it is added to win
	   after each entry in obj. This can be used to add, e.g.,
	   a newline or tab between entries. The separator IS passed
	   recursively.

	   - If obj.nc_wrender is defined, it is assumed to be a function
	   taking a WINDOW handle, and is called. Return value is that of
	   obj.nc_wrender().

	   - As a last resort, it calls obj.toString(). If separator
	   is a string then it is appended to the window as well, as
	   described above for arrays.

	   obj.nc_wrender() implementations should return 0 on success,
	   non-0 on error. They should also not call nc_wrefresh(win).
	*/
	function nc_wrender(win,obj,separator) {
		var dosep = !is_void(separator);
		var rc;
		//print("nc_wrender(",win,",",obj.toSource(),");");
		if( is_object(win) && is_defined( win, 'is_a_JCWindow' ) )
		{
			win = win.window;
		}

		if( ! nc_is_window(win,false) )
		{
			throw new Error("nc_wrender(): first argument must be a WINDOW handle. Got ["+win+"]");
		}
		

		if( is_function(obj) )
		{
			//cerr("nc_wrender() forwarding to functor...");
			rc = obj(win);
		}
		else if( is_object(obj)
			 && is_defined( obj, 'nc_wrender' )
			 )
		{
			//print("Passing on nc_wrender() call. Winid="+win);
			rc = obj.nc_wrender(win);
		}
		else if( is_array( obj ) ) {
			dosep = false;
			for( k in obj ) {
				rc = nc_wrender(win,obj[k],separator);
				if( !rc ) break;
			}
		}
		else
		{
			rc = (0 == nc_waddstr(win,obj));
		}

		if( !rc )
		{
			js_report_warning("nc_wrender("+win+",["+obj.toSource()+"]) rendering failed.");
		}
		else if( dosep )
		{
			rc = nc_waddstr(win,separator);
		}
		return rc;
	}


	/**
	   If (key>=32) or (key==ncurses['ESC']), this function returns 0, otherwise...

	   It strips the control part of the key and returns the "alphabetic"
	   part. e.g., Ctrl-K would return the integer value of 'K'
	   (ncurses['K']). This only works for upper-case letters: we cannot
	   catch Ctrl-Shift-[lowercase_letter] this way.
	*/
	function nc_key_is_ctrl(key) {
		if( key >= 32 ) return 0;
		if( ncurses['ESC'] == key ) return 0;
		key = key & 0x00ff;
		return ncurses['A'] + key - 1;
	}

	/**
	   nc_eval_keymap() looks for a keymap entry named key and evals it (a
	   string or nullary function) if it finds it. If it finds nothing,
	   returns false it returns true. key is expected to be an integer
	   key value, e.g. from nc_wgetch().

	   If the keymap entry is not a string or function then
	   this function throws an exception.

	   Examples of valid keymap entries:

	   km[ascii('a')] = "print('You found the a key.')";

	   km[ascii('A')] = function(){print('You found the A key.');};

	   km[ncurses.F1] = "print('Help not yet implemented.')";

	   km[ncurses.F2] = new String("print('F2 not yet implemented.')");

	**/
	function nc_eval_keymap(key,keymap) {
		var a = keymap[key];
		if( (!a) )
		{ // check against ncurses entry:
			a = keymap[ncurses[key]];
		}
// 		if( (!a) && (key>=32 &&key<=127) )
// 		{ // it might have been entered as an ascii letter
// 			a = keymap[ncurses[key]];
// 		}
		if( ! a )
		{  // unmapped or empty
			
			return false;
		}

		if( is_function(a) )
		{
			try
			{
				a();
				js_run_gc();
			}
			catch(err){
				js_report_error("nc_eval_keymap("+key+") caught an exception:\n"+err);
				return undefined;
			}
			return true;
		}
 		if( a instanceof String ) // 'String' == type_name(a) )
 		{ // kludge
 			eval(a.toString());
 			return true;
 		}
 		if( is_string(a) )
		{
			eval(a);
			return true;
		}


		throw new Error("Unknown action type for keymap entry: "+key);
	}


	/**
	   Creates an array intended for use with nc_eval_keymap(). The returned
	   object has no key mappings assigned to it, but has child subobjects as
	   described below.

	   The keys of the array are intended to be integer values mapping to
	   curses key values. The ncurses global array can act as a translator
	   between these values and their ascii values.

	   The values of the array are unspecified here: see nc_eval_keymap()
	   for how it interprets them.

	   The returned array has these child objects:

	   keymap.main is intended for non-control/non-meta key mappings.

	   keymap.ctrl holds actions for Ctrl keys. Reminder: use ONLY UPPER-CASE
	   LETTERS as keys in this array. Also, some Ctrl keys are reserved by
	   long-standing conventions: Ctrl-C=SIGINT Ctrl-Z=SIGSTOP, Ctrl-S pauses
	   your app and Ctrl-Q resumes it.  C-M and C-J are both the same as
	   ENTER.

	   keymap.meta is also created, intended to hold meta-key (alt/esc)
	   mappings.

	   Note that the returned value is of type Object, not Array, because
	   toSource() does not appear to work as desired (or how i'd expect)
	   on Arrays.
	**/
	function nc_create_keymap()
	{
		var k = new Object();
		k.main = new Object();
		k.ctrl = new Object();
		k.meta = new Object();
		return k;
	}


	/**
	   Returns ch stripped of any curses attributes and whatnot from the
	   character: all bits aside from the bottom 8 are stripped. ch must be a
	   curses-attributed int. It is intended only for reading "simple"
	   characters: those in the ASCII range [32..126].
	**/
	function nc_strip_attr(ch) {
		if( ch < 127 ) return ch;
		return (ch & 0x0F);
	}


	/**
	   Tries to move an NCWindow to a named position, using one of the following
	   patterns:

	   Top or Bottom (move Y only), Left or Right (move X only), Center (X/Y
	   move). Or any pattern from those three groups. e.g. TopCenter,
	   BottomRight. If you pass a "senseless" pattern, like "LeftRight", the
	   order in which Left and Right will be evaluated is undefined.

	   The pos argument must be a string form of one of the known positions.
	   If win has a parent then win is not moved, because this often leads to
	   Grief in curses.

	   Returns true on success, false on error.

	************************************************************************/
	function nc_move_win_to(win,pos) {

		if( win.parent() || ! is_string(pos)) {
			return false;
		}

		if( is_object(win) && is_defined(win,'is_a_JCWindow') ) {
			win = win.window;
		}

		var y = nc_wbegy(win);
		var x = nc_wbegx(win);
		var w = nc_wwidth(win);
		var h = nc_wheight(win);

		var top = 0;

		var sw = nc_screen_width();
		var sh = nc_screen_height();

		var cent = pos.match("Center");

		var matched = false;

		if( pos.match("Top") )
		{
			if( cent ) x = sw/2 - (w/2);
			y = top;
			matched = true;
		}
		else if( pos.match("Bottom" ) )
		{
			if( cent ) x = sw/2 - (w/2);
			y = sh - h;
			matched = true;
		}

		if( pos.match( "Left" ) )
		{
			if( cent ) { y = (sh/2) + top - (h/2);}
			x = 0;
			matched = true;
		}
		else if( pos.match( "Right" )  )
		{
			if( cent ) { y = (sh/2) + top - (h/2);}
			x = sw - w;
			matched = true;
		}

		if( ! matched )
		{
			if( cent ) { x = sw/2 - (w/2); y = (sh/2) + top - (h/2) ; }
			else { x = -1; y = -1; }
		}

		ret = false;
		if( (-1 != y) && (-1 != x) )
		{
			ret = (-1 != nc_mvwin(win,y,x));
			nc_update_panels();
			//win.refresh();
			no_doupdate();
		}
		return ret;
	}

	/**
	   JCListItem is a class for holding arbitrary
	   items for lists. It is mainly an experiment
	   for the nc_wrender() function.
	*/
	function JCListItem(value) {
		this.value = value;
		var self = this;
		this.selected = false;
		/** For use with the global nc_wrender() function. */
		this.nc_wrender = function(win) {
// 			var old = nc_winch(win);
// 			var attr = this.selected
// 			? (ncurses.A_UNDERLINE | ncurses.A_BOLD | ncurses.A_REVERSE)
// 			: old;
// Weird: when i use nc_winch() and nc_wattrset(), all of the
// alpha chars in this.value get lowercased! curses!
			var attr = this.selected
			? (ncurses.A_UNDERLINE | ncurses.A_BOLD | ncurses.A_REVERSE)
			: 0;
			nc_wattron(win, attr);
			nc_wrender(win,this.value);
			nc_wattroff( win, attr );
			//nc_wattrset(win, old);
			
			return true;
		};
	};

	/**
	   Not yet implemented. Experimental.
	*/
	function JCSelectionList()
	{
		var self = this;
		this.window = null;
		var argc = arguments.length;
		if( 1 == argc )
		{
			var o = arguments[0];
			if( is_object(o) && is_defined(o,'is_a_JCWindow') )
			{
				this.window = o.window;
			}
			else if( nc_is_window(o,false) )
			{
				this.window = o;
			}
		}

		if( ! this.window )
		{
			throw new Error("JCSelectionList() ctor: unhandled arguments. Got ["+JCWindow.copyargs(arguments).toString()+"]" );
		}

		this.items = new Array();
		this.pos = 0;
		this.rect = {y:nc_wbegy(this.window),
			     x:nc_wbegx(this.window),
			     w:nc_wwidth(this.window),
			     h:nc_wheight(this.window),
			     maxx:nc_wbegy(this.window)+nc_wmaxy(this.window),
			     maxy:nc_wbegx(this.window)+nc_wmaxx(this.window)
		};
		//print('this.rect ==',this.rect.toSource());

		this.add_item = function(val,sel) {
			var i = new JCListItem(val);
			if( ! is_void(sel) ) i.selected = sel;
			return (self.items[self.items.length] = i);
		};

		this.refresh = function() {
			nc_wmove( this.window, 0,0 );
			var rc = 0;
			var ndx = this.pos;
			var max = this.rect.h;
			var done = 0;
			for( ;
			     (done < this.rect.h) && (ndx < self.items.length);
			     ++ndx, ++done )
			{
				var i = self.items[ndx];
				if( ! i ) continue;
				nc_wmove( this.window, done, 0 );
				rc = nc_wrender( this.window, i ); 
				//print('JCSelectionList drawing item [rc=='+rc+']: '+i.value );
			}
			nc_wrefresh( this.window );
			return rc;
		};

		return self;
	}
	JCSelectionList.inherits(JCWindow);

	/**
	   Uses win to look for a keystroke. It analyzes the keystroke
	   for meta/ctrl state and returns null on error or an object
	   with these properties:

	   .key = integer value of key.

	   .meta = true if this key was ALTed/ESCed, else false.

	   .ctrl = true if this looks like it was a CTRLed key, else false.

	   .label = "pretty" string form of the key.

	*/
	function nc_getkey(win) {
		if( is_object(win) && is_defined(win,'is_a_JCWindow') ) {
			win = win.window;
		}
		if( ! nc_is_window( win ) ) {
			throw new Error("nc_getkey(WINDOW) expects a WINDOW as the first argument.");
		}
		var Key = {key:0,meta:false,ctrl:false,label:""};
		var ESC = ncurses['ESC'];
		var key = 0;
		while( true ) {
			try {
				// the vast majority of this code is handling the meta-key lookup
				var escaped = false;
				var prev = key;
				var k2;
				var prefix = "";
				key = nc_wgetch(win);
				nc_cbreak();
				//print("nc_getkey(): key=="+key+"  "+ncurses[key]);
				if( (ESC == prev) &&
					 ((ESC == key) || (-1==key))
					 ) // simply treat as a single ESC. Do a special case here...
				{
					key = ESC;
					prefix = "";
				}
				else if( (ESC == key) && (key != prev)) // might be metad: wait for next keystroke
				{
					nc_halfdelay(3);
					// next nc_getch() will wait 3/10ths of a second for another key or return -1 on timeout
					continue;
				}
				else if( (ESC == prev) && (ESC != key) ) // metad, but not raw ESC
				{
					Key.meta = true;
					prefix = "Meta-";
				}
				else // non-metad
				{
					k2 = nc_key_is_ctrl(key);
					if( k2 ) // CTRL'd key
					{
						Key.ctrl = true;
						key = k2;
						prefix = "Ctrl-";
					}
					else
					{
						prefix = "";
					}
				}
				Key.key = key;
				Key.label = ""+prefix+ncurses[key];
				break;
			}
			catch(err) {
				js_report_warning("nc_getkey() caught exception:\n"+err);
				Key = null;
			}
		}
		return Key;
	};


} // include guard

1; // true value for include()

